/***************************************************************************
 *   Copyright (C) 2008  Antony Martin <martin(dot)antony(at)yahoo(dot)fr> *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 06/05/2008
   derniere modification : 06/05/2008 */

#include <SCE/SCEMinimal.h>

#include <SCE/interface/SCEOctree.h>


/**
 * \file SCEOctree.c
 * \copydoc octree
 * \file SCEOctree.h
 * \copydoc octree
 */

/**
 * \defgroup octree Octree managment
 * \ingroup interface
 * \brief Octree managment
 */

/** @{ */


static void SCE_Octree_Init (SCE_SOctree *tree)
{
    int i;
    for (i=0; i<8; i++)
        tree->child[i] = NULL;
    tree->have_parent = tree->have_children = SCE_FALSE;
    SCE_BoundingBox_Init (&tree->box);
}

/**
 * \brief Creates an octree
 * \returns a newly allocated octree or NULL on error.
 */
SCE_SOctree* SCE_Octree_Create (void)
{
    SCE_SOctree *tree = SCE_malloc (sizeof *tree);
    if (!tree)
        Logger_LogSrc ();
    else
    {
        SCE_Octree_Init (tree);
    }
    return tree;
}
/**
 * \brief Deletes an octree
 * \param tree an octree to delete
 * 
 * This function delete the given octree and detatch it of any eventual child.
 */
void SCE_Octree_Delete (SCE_SOctree *tree)
{
    if (tree)
    {
        int i;
        for (i=0; i<8; i++)
        {
            if (tree->child[i])
                tree->child[i]->have_parent = SCE_FALSE;
        }
        SCE_free (tree);
    }
}
/**
 * \brief Delete an octree and all its children
 * \param tree an octree
 * 
 * This function deletes recursivly an octree.
 * 
 * \see SCE_Octree_Delete()
 */
void SCE_Octree_DeleteRecursive (SCE_SOctree *tree)
{
    if (tree)
    {
        int i;
        for (i=0; i<8; i++)
        {
            SCE_Octree_DeleteRecursive (tree->child[i]);
            tree->child[i] = NULL;
        }
        SCE_Octree_Delete (tree);
    }
}

/**
 * \brief Sets the center point of an octree
 * \param tree an octree
 * \param x,y,z the new center's coordinates
 * 
 * \see SCE_Octree_SetCenterv()
 */
void SCE_Octree_SetCenter (SCE_SOctree *tree, float x, float y, float z)
{
    SCE_BoundingBox_SetCenter (&tree->box, x, y, z);
}
/**
 * \brief Sets the center point of an octree
 * \param tree an octree
 * \param c vector of the new center's coordinates
 * 
 * \see SCE_Octree_SetCenter()
 */
void SCE_Octree_SetCenterv (SCE_SOctree *tree, SCE_TVector3 c)
{
    SCE_BoundingBox_SetCenterv (&tree->box, c);
}
/**
 * \brief Gets the center point of an octree
 * \param tree an octree
 * \param c vector to store center's coordinates
 */
void SCE_Octree_GetCenterv (SCE_SOctree *tree, SCE_TVector3 c)
{
    SCE_BoundingBox_GetCenterv (&tree->box, c);
}
/**
 * \brief Sets the size of an octree
 * \param tree an octree
 * \param w,h,d new width, height and depth of the octree
 */
void SCE_Octree_SetSize (SCE_SOctree *tree, float w, float h, float d)
{
    SCE_BoundingBox_SetSize (&tree->box, w, h, d);
}
/**
 * \brief Gets the width of an octree
 * \param tree an octree
 * \returns the width of the given octree
 * 
 * \see SCE_Octree_GetHeight()
 * \see SCE_Octree_GetDepth()
 */
float SCE_Octree_GetWidth (SCE_SOctree *tree)
{
    return SCE_BoundingBox_GetWidth (&tree->box);
}
/**
 * \brief Gets the height of an octree
 * \param tree an octree
 * \returns the height of the given octree
 * 
 * \see SCE_Octree_GetWidth()
 * \see SCE_Octree_GetDepth()
 */
float SCE_Octree_GetHeight (SCE_SOctree *tree)
{
    return SCE_BoundingBox_GetHeight (&tree->box);
}
/**
 * \brief Gets the depth of an octree
 * \param tree an octree
 * \returns the depth of the given octree
 * 
 * \see SCE_Octree_GetWidth()
 * \see SCE_Octree_GetHeight()
 */
float SCE_Octree_GetDepth (SCE_SOctree *tree)
{
    return SCE_BoundingBox_GetDepth (&tree->box);
}

/**
 * \brief Gets the chlidren of an octree
 * \param tree an octree
 * \return an array of the given octree's chlidren
 */
SCE_SOctree** SCE_Octree_GetChildren (SCE_SOctree *tree)
{
    return tree->child;
}

/**
 * \brief Creates the eight chlidren of an octree
 * \param tree an octree
 * \returns SCE_ERROR on error, SCE_OK otherwise
 * 
 * This function creates the eight children of the given octree.
 */
int SCE_Octree_MakeChildren (SCE_SOctree *tree)
{
    int i;
    SCE_TVector3 center, centers[8], tmp;
    float w = SCE_BoundingBox_GetWidth (&tree->box);
    float h = SCE_BoundingBox_GetHeight (&tree->box);
    float d = SCE_BoundingBox_GetDepth (&tree->box);
    SCE_BoundingBox_GetCenterv (&tree->box, center);

    tmp[2] = d/4.0f;
    for (i=0; i<8; i+=4)
    {
        tmp[0] =  w/4.0f; tmp[1] =  h/4.0f;
        SCE_Vector3_Operator2v (centers[i], =, center, +, tmp);
        tmp[0] = -w/4.0f; tmp[1] =  h/4.0f;
        SCE_Vector3_Operator2v (centers[i+1], =, center, +, tmp);
        tmp[0] = -w/4.0f; tmp[1] = -h/4.0f;
        SCE_Vector3_Operator2v (centers[i+2], =, center, +, tmp);
        tmp[0] =  w/4.0f; tmp[1] = -h/4.0f;
        SCE_Vector3_Operator2v (centers[i+3], =, center, +, tmp);
        tmp[2] = -d/4.0f;
    }

    for (i=0; i<8; i++)
    {
        tree->child[i] = SCE_Octree_Create ();
        if (!tree->child[i])
        {
            Logger_LogSrc ();
            return SCE_ERROR;
        }
        SCE_BoundingBox_SetFromCenter (&tree->child[i]->box, centers[i],
                                       w/2.0f, h/2.0f, d/2.0f);
    }
    return SCE_OK;
}

/**
 * \brief Creates octree's clidren recursivly
 * \param tree an octree
 * \param stop a function to determines when stop chlidren generation
 * \param param user data to pass to the \p stop function
 * \returns SCE_ERROR on error, SCE_OK otherwise
 * 
 * This function creates octress's chlidren recursivly. The recursion depth is
 * controlled by the \p stop function.
 * 
 * \see SCE_FOctreeLimitFunc
 * \see SCE_Octree_MakeChildren()
 */
int SCE_Octree_RecursiveMake (SCE_SOctree *tree, SCE_FOctreeLimitFunc stop,
                              void *param)
{
    if (!stop (tree, param))
    {
        int i;
        if (SCE_Octree_MakeChildren (tree) < 0)
        {
            Logger_LogSrc ();
            return SCE_ERROR;
        }
        for (i=0; i<8; i++)
        {
            if (SCE_Octree_RecursiveMake (tree->child[i], stop, param) < 0)
            {
                Logger_LogSrc ();
                return SCE_ERROR;
            }
        }
    }
    return SCE_OK;
}


/** @} */
